#pragma once

/* Icons for places */
#define NAUTILUS_ICON_FILESYSTEM	"drive-harddisk-symbolic"
#define NAUTILUS_ICON_FOLDER		"folder-symbolic"
#define NAUTILUS_ICON_FOLDER_REMOTE	"folder-remote-symbolic"
#define NAUTILUS_ICON_HOME		"user-home-symbolic"

#define NAUTILUS_ICON_FOLDER_DOCUMENTS  "folder-documents-symbolic"
#define NAUTILUS_ICON_FOLDER_DOWNLOAD   "folder-download-symbolic"
#define NAUTILUS_ICON_FOLDER_MUSIC      "folder-music-symbolic"
#define NAUTILUS_ICON_FOLDER_PICTURES   "folder-pictures-symbolic"
#define NAUTILUS_ICON_FOLDER_PUBLIC_SHARE "folder-publicshare-symbolic"
#define NAUTILUS_ICON_FOLDER_TEMPLATES  "folder-templates-symbolic"
#define NAUTILUS_ICON_FOLDER_VIDEOS     "folder-videos-symbolic"

/* Fullcolor icons */
#define NAUTILUS_ICON_FULLCOLOR_FOLDER              "folder"
#define NAUTILUS_ICON_FULLCOLOR_FOLDER_REMOTE       "folder-remote"
#define NAUTILUS_ICON_FULLCOLOR_FOLDER_DOCUMENTS    "folder-documents"
#define NAUTILUS_ICON_FULLCOLOR_FOLDER_DOWNLOAD     "folder-download"
#define NAUTILUS_ICON_FULLCOLOR_FOLDER_MUSIC        "folder-music"
#define NAUTILUS_ICON_FULLCOLOR_FOLDER_PICTURES     "folder-pictures"
#define NAUTILUS_ICON_FULLCOLOR_FOLDER_PUBLIC_SHARE "folder-publicshare"
#define NAUTILUS_ICON_FULLCOLOR_FOLDER_TEMPLATES    "folder-templates"
#define NAUTILUS_ICON_FULLCOLOR_FOLDER_VIDEOS       "folder-videos"